import logging
logger = logging.getLogger('isitfit')

import pandas as pd
from tabulate import tabulate

# https://pypi.org/project/termcolor/
from termcolor import colored


class UtilizationListener:

  def __init__(self):
    # iterate over all ec2 instances
    self.sum_capacity = 0
    self.sum_used = 0
    self.df_all = []
    self.table = None # will contain the final table after calling `after_all`


  def per_ec2(self, ec2_obj, ec2_df, mm, ddg_df):
    """
    Listener function to be called upon the download of each EC2 instance's data
    ec2_obj - boto3 resource
    ec2_df - pandas dataframe with data from cloudwatch+cloudtrail
    mm - mainManager class
    ddg_df - dataframe of data from datadog: {cpu,ram}-{max,avg}
    """
    # results: 2 numbers: capacity (USD), used (USD)
    res_capacity = (ec2_df.nhours*ec2_df.cost_hourly).sum()

    if 'ram_used_avg.datadog' in ec2_df.columns:
      # use both the CPU Average from cloudwatch and the RAM average from datadog
      utilization_factor = ec2_df[['Average', 'ram_used_avg.datadog']].mean(axis=1, skipna=True)
    else:
      # use only the CPU average from cloudwatch
      utilization_factor = ec2_df.Average

    res_used     = (ec2_df.nhours*ec2_df.cost_hourly*utilization_factor/100).sum()
    #logger.debug("res_capacity=%s, res_used=%s"%(res_capacity, res_used))

    self.sum_capacity += res_capacity
    self.sum_used += res_used
    self.df_all.append({'instance_id': ec2_obj.instance_id, 'capacity': res_capacity, 'used': res_used})


  def after_all(self, n_ec2_total, mm, n_ec2_analysed):
    # for debugging
    df_all = pd.DataFrame(self.df_all)
    logger.debug("\ncapacity/used per instance")
    logger.debug(df_all)
    logger.debug("\n")

    cwau_val = 0
    if self.sum_capacity!=0:
      cwau_val = self.sum_used/self.sum_capacity*100

    cwau_color = 'orange'
    if cwau_val >= 70:
      cwau_color = 'green'
    elif cwau_val <= 30:
      cwau_color = 'red'

    dt_start = mm.StartTime.strftime("%Y-%m-%d")
    dt_end   = mm.EndTime.strftime("%Y-%m-%d")
    
    self.table = [
      ["Start date", "%s"%dt_start],
      ["End date", "%s"%dt_end],
      ["EC2 machines (total)", "%i"%n_ec2_total],
      ["EC2 machines (analysed)", "%i"%n_ec2_analysed],
      [colored("Billed cost", 'cyan'), colored("%0.0f $"%self.sum_capacity, 'cyan')],
      [colored("Used cost", 'cyan'), colored("%0.0f $"%self.sum_used, 'cyan')],
      [colored("CWAU (Used/Billed)", cwau_color), colored("%0.0f %%"%cwau_val, cwau_color)],
    ]


  def display_all(self, *args, **kwargs):
    # logger.info("Summary:")
    logger.info("Cost-Weighted Average Utilization (CWAU) of the AWS EC2 account:")
    logger.info("")
    logger.info(tabulate(self.table, headers=['Field', 'Value']))
    logger.info("")
    logger.info("For reference:")
    logger.info(colored("* CWAU >= 70% is well optimized", 'green'))
    logger.info(colored("* CWAU <= 30% is underused", 'red'))




import logging
logger = logging.getLogger('isitfit')

class RedisPandas:
  """
  Python class that manages caching pandas dataframes to redis
  https://stackoverflow.com/a/57986261/4126114
  """
  def __init__(self):
    self.redis_args = {}
    self.redis_client = None
    self.pyarrow_context = None
    
  def fetch_envvars(self):
    # check redis parameters if set for caching
    import os
    for k1, k2 in [
        ('host', "ISITFIT_REDIS_HOST"),
        ('port', "ISITFIT_REDIS_PORT"),
        ('db', "ISITFIT_REDIS_DB")
      ]:
      self.redis_args[k1] = os.getenv(k2, None)

  def isSetup(self):
    return any(self.redis_args.values())

  def connect(self):
    logger.info("Connecting to redis cache")
    import redis
    import pyarrow as pa

    self.redis_client = redis.Redis(**self.redis_args)
    self.pyarrow_context = pa.default_serialization_context()

  def isReady(self):
    return self.redis_client is not None

  def set(self, key, df):
    pybytes = self.pyarrow_context.serialize(df).to_buffer().to_pybytes()
    self.redis_client.set(key, pybytes)

  def get(self, key):
    v1 = self.redis_client.get(key)
    if not v1: return v1
    v2 = self.pyarrow_context.deserialize(v1)
    return v2



import os

class DotMan:
  def get_dotisitfit(self):
    # get home
    import pathlib
    p1_home = str(pathlib.Path.home())

    # check dot folder
    p2_dot = os.path.join(p1_home, ".isitfit")
    if not os.path.exists(p2_dot):
      pathlib.Path(p2_dot).mkdir(exist_ok=True)

    return p2_dot


  def get_myuid(self, is_reentry=False):
    """
    Create a UUID for each installation of isitfit
    This also creates a .isitfit folder in the user's home directory
    and caches the generated UUID in a txt file for re-use

    is_reentry - internally used flag to identify that this is a case when
                 UUID is identified as invalid and needs to be set again
    """
    p2_dot = self.get_dotisitfit()

    # check uid file within dot folder
    p3_uidtxt = os.path.join(p2_dot, "uid.txt")
    uuid_val = None
    if not os.path.exists(p3_uidtxt):
      import uuid
      uuid_val = uuid.uuid4().hex
      with open(p3_uidtxt, 'w') as fh:
        fh.write(uuid_val)

    # if not created above, read from file
    if uuid_val is None:
      with open(p3_uidtxt, 'r') as fh:
        uuid_val = fh.read()

    # if re-entry due to invalid ID or not
    if is_reentry:
      # any further processing of this would be an overkill
      pass
    else:
      # verify that the UUID is valid (in case of accidental overwrite)
      if len(uuid_val)!=32:
        # drop the uid.txt file and overwrite it
        os.remove(p3_uidtxt)
        uuid_val = self.get_myuid(True)

    # return
    return uuid_val




import logging
logger = logging.getLogger('isitfit')

from .tagsSuggestBasic import TagsSuggestBasic
from .utils import MAX_ROWS
import os
import requests
import json
from .utils import IsitfitError
import urllib.parse

#BASE_URL = 'https://r0ju8gtgtk.execute-api.us-east-1.amazonaws.com/dev/'
BASE_URL = 'https://api.isitfit.io/v0/'

class TagsSuggestAdvanced(TagsSuggestBasic):

  def __init__(self):
    logger.debug("TagsSuggestAdvanced::constructor")

    # boto3 ec2 and cloudwatch data
    import boto3
    self.sts = boto3.client('sts')
    self.sqs_res = boto3.resource('sqs', region_name='us-east-1') # region matches with the serverless.yml region
    self.s3_client  = boto3.client('s3' )
    return super().__init__()


  def prepare(self):
    logger.debug("TagsSuggestAdvanced::prepare")

    self._register()
    if not 'status' in self.r_register:
      raise IsitfitError("Failed to ping the remote: %s"%self.r_register)

    # TODO implement later
    # print(self.r_register)
    logger.debug("Granted access to s3 arn: %s"%self.r_register['s3_arn'])
    logger.debug("Granted access to sqs url: %s"%self.r_register['sqs_url'])
    logger.debug("Note that account number 974668457921 is AutofitCloud, the company behind isitfit.")
    logger.debug("For more info, visit https://autofitcloud.com/privacy")

    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/sqs.html#SQS.Queue.receive_messages
    self.sqs_q = self.sqs_res.Queue(self.r_register['sqs_url'])


  def suggest(self):
    logger.debug("TagsSuggestAdvanced::suggest")

    import tempfile

    logger.info("Uploading ec2 names to s3")
    with tempfile.NamedTemporaryFile(suffix='.csv', prefix='isitfit-ec2names-', delete=True) as fh:
      logger.debug("Will use temporary file %s"%fh.name)
      self.tags_df.to_csv(fh.name, index=False)
      self.s3_key_suffix = 'tags_request.csv'
      s3_path = os.path.join(self.r_sts['Account'], self.r_sts['UserId'], self.s3_key_suffix)
      self.s3_client.put_object(Bucket=self.r_register['s3_bucketName'], Key=s3_path, Body=fh)

    # mark timestamp of request
    import datetime as dt
    dt_now = dt.datetime.utcnow() #.strftime('%s')

    # POST /tags/suggest
    self._tags_suggest()

    # now listen on sqs
    # https://github.com/jegesh/python-sqs-listener/blob/master/sqs_listener/__init__.py#L123
    logger.info("Waiting for results")
    MAX_RETRIES = 5
    i_retry = 0
    any_found = False
    import time
    n_secs = 5
    while i_retry < MAX_RETRIES:
      i_retry += 1

      if i_retry == 1:
        time.sleep(1)
      else:
        #logger.info("Sleep %i seconds"%n_secs)
        time.sleep(n_secs)

      logger.debug("Check sqs messages (Retry %i/%i)"%(i_retry, MAX_RETRIES))
      messages = self.sqs_q.receive_messages(
        AttributeNames=['SentTimestamp'],
        QueueUrl=self.sqs_q.url,
        MaxNumberOfMessages=10
      )
      logger.debug("{} messages received".format(len(messages)))
      import datetime as dt
      for m in messages:
          any_found = True
          sentTime_dt = None
          sentTime_str = "-"
          if m.attributes is not None:
            sentTime_dt = m.attributes['SentTimestamp']
            sentTime_dt = dt.datetime.utcfromtimestamp(int(sentTime_dt)/1000)
            sentTime_str = sentTime_dt.strftime("%Y/%m/%d %H:%M:%S")

          logger.debug("Message: %s: %s"%(sentTime_str, m.body))

          try:
            m.body_decoded = json.loads(m.body)
          except json.decoder.JSONDecodeError as e:
            logger.debug("(Invalid message with non-json body. Skipping)")
            continue

          if 'type' not in m.body_decoded:
            # print("FOOOOOOOOOO")
            logger.debug("(Message body missing key 'type'. Skipping)")
            continue

          if m.body_decoded['type'] != 'tags suggest':
            logger.debug("(Message topic = %s != tags suggest. Skipping)")
            continue

          if (sentTime_dt < dt_now):
              logger.debug("(Stale message. Dropping and skipping)")
              m.delete()
              continue

          # all "tags suggest" messages are removed from the queue
          logger.debug("(Message is ok. Will process. Removing from queue)")
          m.delete()

          # process messages
          logger.info("Server message: %s"%m.body_decoded['status'])
          if m.body_decoded['status'] != 'calculation complete':
            continue

          if m.body_decoded['status'] == 'calculation complete':
            # upon calculation complete message
            if 's3_key_suffix' not in m.body_decoded:
              logger.debug("(Missing s3_key_suffix key from body. Aborting)")
              return

            self.csv_fn = None
            with tempfile.NamedTemporaryFile(suffix='.csv', prefix='isitfit-tags-suggestAdvanced-', delete=False) as fh:
              self.csv_fn = fh.name
              s3_path = os.path.join(self.r_register['s3_keyPrefix'], m.body_decoded['s3_key_suffix'])
              logger.info("Downloading tag suggestions from isitfit server")
              logger.debug("Getting s3 file %s"%s3_path)
              logger.debug("Saving it into %s"%fh.name)
              response = self.s3_client.get_object(Bucket=self.r_register['s3_bucketName'], Key=s3_path)
              fh.write(response['Body'].read())

            logger.debug("TagsSuggestAdvanced:suggest .. read_csv")
            import pandas as pd
            self.suggested_df = pd.read_csv(self.csv_fn, nrows=MAX_ROWS)

            # count number of rows in csv
            # https://stackoverflow.com/a/36973958/4126114
            logger.debug("TagsSuggestAdvanced:suggest .. count_rows")
            with open(fh.name) as f2:
                self.suggested_shape = [sum(1 for line in f2), 4] # 4 is just hardcoded number of columns that doesn't matter much

            logger.debug("TagsSuggestAdvanced:suggest .. done")
            return

    # if nothing returned on sqs
    if not any_found:
      logger.error("Absolute radio silence on sqs :(")

    # either no sqs messages,
    # or found some sqs messages, but none were for tags request fulfilled
    import pandas as pd
    self.suggested_df = pd.DataFrame()
    self.suggested_shape = [0,4]
    self.csv_fn = None


  def _register(self):
      logger.info("Logging into server")

      # https://stackoverflow.com/a/8223955/4126114
      URL = urllib.parse.urljoin(BASE_URL, './register')
      logger.debug("POST %s"%URL)

      self.r_sts = self.sts.get_caller_identity()
      del self.r_sts['ResponseMetadata']
      # eg {'UserId': 'AIDA6F3WEM7AXY6Y4VWDC', 'Account': '974668457921', 'Arn': 'arn:aws:iam::974668457921:user/shadi'}

      r1 = requests.request('post', URL, json=self.r_sts)

      # https://stackoverflow.com/questions/18810777/how-do-i-read-a-response-from-python-requests
      self.r_register = json.loads(r1.text)

      # check for internal server error
      if 'message' in self.r_register:
        if self.r_register['message']=='Internal server error':
          raise IsitfitError('Internal server error')

      # check schema
      from schema import Schema, Optional, SchemaError
      register_schema = Schema({
        'status': str,
        's3_arn': str,
        's3_bucketName': str,
        's3_keyPrefix': str,
        'sqs_url': str,
        Optional(str): object
      })
      try:
        register_schema.validate(self.r_register)
      except SchemaError as e:
        logger.error("Received response: %s"%r1.text)
        raise IsitfitError("Invalid register response: %s"%str(e))



  def _tags_suggest(self):
      logger.info("Requesting tag suggestions from isitfit server")

      # https://stackoverflow.com/a/8223955/4126114
      URL = urllib.parse.urljoin(BASE_URL, './tags/suggest')
      logger.debug("POST %s"%URL)

      load_send = {}
      load_send.update(self.r_sts)
      load_send['s3_key_suffix'] = self.s3_key_suffix
      load_send['sqs_url'] = self.r_register['sqs_url']
      r_tagsSuggest = requests.request('post', URL, json=load_send)
      # https://stackoverflow.com/questions/18810777/how-do-i-read-a-response-from-python-requests
      r2 = json.loads(r_tagsSuggest.text)

      if 'error' in r2:
        print(r2)
        raise IsitfitError('Serverside error: %s'%r2['error'])

      if 'message' in r2:
        print(r2)
        raise IsitfitError('Serverside error: %s'%r2['message'])

      return r2



import boto3
import pandas as pd
from tqdm import tqdm
import datetime as dt
import numpy as np
import pytz
from .cacheManager import RedisPandas as RedisPandasCacheManager
from .datadogManager import HostNotFoundInDdg, DataNotFoundForHostInDdg

import logging
logger = logging.getLogger('isitfit')


from .utils import mergeSeriesOnTimestampRange, ec2_catalog
from .cloudtrail_ec2type import Manager as CloudtrailEc2typeManager


SECONDS_IN_ONE_DAY = 60*60*24 # 86400  # used for granularity (daily)
MINUTES_IN_ONE_DAY = 60*24 # 1440
N_DAYS=90

from .utils import IsitfitError

class NoCloudtrailException(Exception):
    pass

class NoCloudwatchException(Exception):
    pass


def myreturn(df_xxx):
    if df_xxx.shape[0] > 0:
      return df_xxx
    else:
      return None # this means that the data was found in cache, but it was empty (meaning aws returned no data)


def tagsContain(f_tn, ec2_obj):
  for t in ec2_obj.tags:
    for k in ['Key', 'Value']:
      if f_tn in t[k].lower():
        return True

  return False


class MainManager:
    def __init__(self, ddg=None, filter_tags=None):
        # boto3 ec2 and cloudwatch data
        self.ec2_resource = boto3.resource('ec2')
        self.cloudwatch_resource = boto3.resource('cloudwatch')

        # set start/end dates
        dt_now_d=dt.datetime.now().replace(tzinfo=pytz.utc)
        self.StartTime=dt_now_d - dt.timedelta(days=N_DAYS)
        self.EndTime=dt_now_d
        logger.debug("Metrics start..end: %s .. %s"%(self.StartTime, self.EndTime))

        # manager of redis-pandas caching
        self.cache_man = RedisPandasCacheManager()

        # boto3 cloudtrail data
        cloudtrail_client = boto3.client('cloudtrail')
        self.cloudtrail_manager = CloudtrailEc2typeManager(cloudtrail_client, dt_now_d, self.cache_man)

        # listeners post ec2 data fetch and post all activities
        self.listeners = {'pre':[], 'ec2': [], 'all': []}

        # datadog manager
        self.ddg = ddg

        # filtering by tags
        self.filter_tags = filter_tags


    def add_listener(self, event, listener):
      if event not in self.listeners:
        raise IsitfitError("Internal dev error: Event %s is not supported for listeners. Use: %s"%(event, ",".join(self.listeners.keys())))

      self.listeners[event].append(listener)


    def get_ifi(self):
        # set up caching if requested
        self.cache_man.fetch_envvars()
        if self.cache_man.isSetup():
          self.cache_man.connect()

        # 0th pass to count
        n_ec2_total = len(list(self.ec2_resource.instances.all()))
        logger.warning("Found a total of %i EC2 instances"%n_ec2_total)

        if n_ec2_total==0:
          return

        # call listeners
        for l in self.listeners['pre']:
          l()

        # download ec2 catalog: 2 columns: ec2 type, ec2 cost per hour
        self.df_cat = ec2_catalog()

        # get cloudtail ec2 type changes for all instances
        self.cloudtrail_manager.init_data(self.ec2_resource.instances.all(), n_ec2_total)

        # iterate over all ec2 instances
        n_ec2_analysed = 0
        sum_capacity = 0
        sum_used = 0
        df_all = []
        ec2_noCloudwatch = []
        ec2_noCloudtrail = []
        for ec2_obj in tqdm(self.ec2_resource.instances.all(), total=n_ec2_total, desc="Second pass through EC2 instances", initial=1):
          # if filters requested, check that this instance passes
          if self.filter_tags is not None:
            f_tn = self.filter_tags.lower()
            passesFilter = tagsContain(f_tn, ec2_obj)
            if not passesFilter:
              continue

          try:
            ec2_df, ddg_df = self._handle_ec2obj(ec2_obj)
            n_ec2_analysed += 1

            # call listeners
            for l in self.listeners['ec2']:
              l(ec2_obj, ec2_df, self, ddg_df)

          except NoCloudwatchException:
            ec2_noCloudwatch.append(ec2_obj.instance_id)
          except NoCloudtrailException:
            ec2_noCloudtrail.append(ec2_obj.instance_id)

        # call listeners
        logger.info("... done")
        logger.info("")
        logger.info("")

        if len(ec2_noCloudwatch)>0:
          has_more_cw = "..." if len(ec2_noCloudwatch)>5 else ""
          l_no_cw = ", ".join(ec2_noCloudwatch[:5])
          logger.warning("No cloudwatch data for: %s%s"%(l_no_cw, has_more_cw))
          logger.info("")

        if len(ec2_noCloudtrail)>0:
          has_more_ct = "..." if len(ec2_noCloudtrail)>5 else ""
          l_no_ct = ", ".join(ec2_noCloudtrail[:5])
          logger.warning("No cloudtrail data for: %s%s"%(l_no_ct, has_more_ct))
          logger.info("")

        for l in self.listeners['all']:
          l(n_ec2_total, self, n_ec2_analysed)

        logger.info("")
        logger.info("")
        return


    def _cloudwatch_metrics_cached(self, ec2_obj):

        # check cache first
        cache_key = "mainManager._cloudwatch_metrics/%s"%ec2_obj.instance_id
        if self.cache_man.isReady():
          df_cache = self.cache_man.get(cache_key)
          if df_cache is not None:
            logger.debug("Found cloudwatch metrics in redis cache for %s, and data.shape[0] = %i"%(ec2_obj.instance_id, df_cache.shape[0]))
            return myreturn(df_cache)

        # if no cache, then download
        df_fresh = self._cloudwatch_metrics_core(ec2_obj)

        # if caching enabled, store it for later fetching
        # https://stackoverflow.com/a/57986261/4126114
        # Note that this even stores the result if it was "None" (meaning that no data was found)
        if self.cache_man.isReady():
          self.cache_man.set(cache_key, df_fresh)

        # done
        return myreturn(df_fresh)


    def _cloudwatch_metrics_core(self, ec2_obj):
        """
        Return a pandas series of CPU utilization, daily max, for 90 days
        """
        metrics_iterator = self.cloudwatch_resource.metrics.filter(
            Namespace='AWS/EC2', 
            MetricName='CPUUtilization', 
            Dimensions=[{'Name': 'InstanceId', 'Value': ec2_obj.instance_id}]
          )
        df_cw1 = []
        for m_i in metrics_iterator:
            json_i = m_i.get_statistics(
              Dimensions=[{'Name': 'InstanceId', 'Value': ec2_obj.instance_id}],
              Period=SECONDS_IN_ONE_DAY,
              Statistics=['Average', 'SampleCount', 'Maximum'],
              Unit='Percent',
              StartTime=self.StartTime,
              EndTime=self.EndTime
            )
            # logger.debug(json_i)
            if len(json_i['Datapoints'])==0: continue # skip (no data)

            df_i = pd.DataFrame(json_i['Datapoints'])

            # edit 2019-09-13: no need to subsample columns
            # The initial goal was to drop the "Unit" column (which just said "Percent"),
            # but it's not such a big deal, and avoiding this subsampling simplifies the code a bit
            # df_i = df_i[['Timestamp', 'SampleCount', 'Average']]

            # sort and append in case of multiple metrics
            df_i = df_i.sort_values(['Timestamp'], ascending=True)
            df_cw1.append(df_i)

        if len(df_cw1)==0:
          return pd.DataFrame() # use an empty dataframe in order to distinguish when getting from cache if not available in cache or data not found but set in cache

        if len(df_cw1) >1:
          raise IsitfitError("More than 1 cloudwatch metric found for %s"%ec2_obj.instance_id)

        # merge
        # df_cw2 = pd.concat(df_cw1, axis=1)

        # dataframe to be returned
        df_cw3 = df_cw1[0]

        # before returning, convert dateutil timezone to pytz
        # for https://github.com/pandas-dev/pandas/issues/25423
        # via https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.Series.dt.tz_convert.html
        # Edit 2019-09-25 Instead of keeping the full timestamp, just truncate to date, especially that this is just daily data
        # df_cw3['Timestamp'] = df_cw3.Timestamp.dt.tz_convert(pytz.utc)
        df_cw3['Timestamp'] = df_cw3.Timestamp.dt.date

        # done
        return df_cw3


    def _get_ddg_cached(self, ec2_obj):
        # check if we can get datadog data
        if not self.ddg:
          return None

        if not self.ddg.is_configured():
          return None

        # check cache first
        cache_key = "mainManager._get_ddg_cached/%s"%ec2_obj.instance_id
        if self.cache_man.isReady():
          df_cache = self.cache_man.get(cache_key)
          if df_cache is not None:
            logger.debug("Found datadog metrics in redis cache for %s, and data.shape[0] = %i"%(ec2_obj.instance_id, df_cache.shape[0]))
            return myreturn(df_cache)

        # if no cache, then download
        df_fresh = pd.DataFrame() # use an empty dataframe in order to distinguish when getting from cache if not available in cache or data not found but set in cache
        try:
          df_fresh = self.ddg.get_metrics_all(ec2_obj.instance_id)
        except HostNotFoundInDdg:
          pass
        except DataNotFoundForHostInDdg:
          pass

        # if caching enabled, store it for later fetching
        # https://stackoverflow.com/a/57986261/4126114
        # Note that this even stores the result if it was "None" (meaning that no data was found)
        if self.cache_man.isReady():
          # print("Saving to redis %s"%ec2_obj.instance_id)
          self.cache_man.set(cache_key, df_fresh)

        # done
        return myreturn(df_fresh)


    def _handle_ec2obj(self, ec2_obj):
        # logger.debug("%s, %s"%(ec2_obj.instance_id, ec2_obj.instance_type))

        # pandas series of CPU utilization, daily max, for 90 days
        df_metrics = self._cloudwatch_metrics_cached(ec2_obj)

        # no data
        if df_metrics is None:
          raise NoCloudwatchException("No cloudwatch data for %s"%ec2_obj.instance_id)

        # pandas series of number of cpu's available on the machine over time, past 90 days
        df_type_ts1 = self.cloudtrail_manager.single(ec2_obj)
        if df_type_ts1 is None:
          raise NoCloudtrailException("No cloudtrail data for %s"%ec2_obj.instance_id)

        # this is redundant with the implementation in _cloudwatch_metrics_core,
        # and it's here just in case the cached redis version is not a date,
        # but it's not really worth it to make a full refresh of the cache for this
        # if df_metrics.Timestamp.dtype==dt.date:
        try:
          df_metrics['Timestamp'] = df_metrics.Timestamp.dt.date
        except AttributeError:
          pass

        # convert type timeseries to the same timeframes as pcpu and n5mn
        #if ec2_obj.instance_id=='i-069a7808addd143c7':
        ec2_df = mergeSeriesOnTimestampRange(df_metrics, df_type_ts1)
        #logger.debug("\nafter merge series on timestamp range")
        #logger.debug(ec2_df.head())

        # merge with type changes (can't use .merge on timestamps, need to use .concat)
        #ec2_df = df_metrics.merge(df_type_ts2, left_on='Timestamp', right_on='EventTime', how='left')
        # ec2_df = pd.concat([df_metrics, df_type_ts2], axis=1)

        # merge with catalog
        ec2_df = ec2_df.merge(self.df_cat[['API Name', 'cost_hourly']], left_on='instanceType', right_on='API Name', how='left')
        #logger.debug("\nafter merge with catalog")
        #logger.debug(ec2_df.head())

        # calculate number of running hours
        # In the latest 90 days, sampling is per minute in cloudwatch
        # https://aws.amazon.com/cloudwatch/faqs/
        # Q: What is the minimum resolution for the data that Amazon CloudWatch receives and aggregates?
        # A: ... For example, if you request for 1-minute data for a day from 10 days ago, you will receive the 1440 data points ...
        ec2_df['nhours'] = np.ceil(ec2_df.SampleCount/60)

        # check if we can get datadog data
        ddg_df = self._get_ddg_cached(ec2_obj)
        # print("ddg data", ddg_df)

        if ddg_df is not None:
          # convert from datetime to date to be able to merge with ec2_df
          ddg_df['ts_dt'] = ddg_df.ts_dt.dt.date
          # append the datadog suffix
          ddg_df = ddg_df.add_suffix('.datadog')
          # merge
          ec2_df = ec2_df.merge(ddg_df, how='outer', left_on='Timestamp', right_on='ts_dt.datadog')

        return ec2_df, ddg_df



# RuntimeError: Click will abort further execution because Python 3 was configured to use ASCII as encoding for the environment. 
# Consult https://click.palletsprojects.com/en/7.x/python3/ for mitigation steps.
# 
# Edit 2019-10-08: whatsapp's wadebug uses "click.disable_unicode_literals_warning = True"
# Ref: https://github.com/WhatsApp/WADebug/blob/958ac37be804cc732ae514d4872b93d19d197a5c/wadebug/cli.py#L23
from .utils import mysetlocale
mysetlocale()


import logging
logger = logging.getLogger('isitfit')

import click

from . import isitfit_version

def display_footer():
    logger.info("")
    logger.info("Generated by isitfit version %s"%isitfit_version)
    logger.info("For more info about isitfit, check https://isitfit.autofitcloud.com ⛅")
    logger.info("Also, consider following the Global Climate Strike news at https://twitter.com/hashtag/ClimateStrike 🌎")

# With atexit, this message is being displayed even in case of early return or errors.
# Changing to try/finally in the __main__ below
#import atexit
#atexit.register(display_footer)


@click.group(invoke_without_command=True)
@click.option('--debug', is_flag=True, help='Display more details to help with debugging')
@click.option('--version', is_flag=True, help='Show the installed version')
@click.option('--optimize', is_flag=True, help='Generate recommendations of optimal EC2 sizes')
@click.option('--n', default=0, help='number of underused ec2 optimizations to find before stopping. Skip to get all optimizations')
@click.option('--filter-tags', default=None, help='filter instances for only those carrying this value in the tag name or value')
@click.pass_context
def cli(ctx, debug, version, optimize, n, filter_tags):
    if version:
      print('isitfit version %s'%isitfit_version)
      return

    logLevel = logging.DEBUG if debug else logging.INFO
    ch = logging.StreamHandler()
    ch.setLevel(logLevel)
    logger.addHandler(ch)
    logger.setLevel(logLevel)

    if debug:
      logger.debug("Enabled debug level")
      logger.debug("-------------------")

    # check if current version is out-of-date
    from .utils import prompt_upgrade
    is_outdated = prompt_upgrade('isitfit', isitfit_version)
    if is_outdated:
      # Give the user some time to read the message and possibly update
      import time
      time.sleep(3)

    # do not continue with the remaining code here
    # if a command is invoked, eg `isitfit tags`
    ctx.ensure_object(dict)
    if ctx.invoked_subcommand is not None:
      return

    # gather anonymous usage statistics
    from .utils import ping_matomo
    if optimize:
      ping_matomo("/cost/optimize")
    else:
      ping_matomo("/cost/analyze")

    #logger.info("Is it fit?")
    from .utils import IsitfitError
    try:
      logger.info("Initializing...")

      # moved these imports from outside the function to inside it so that `isitfit --version` wouldn't take 5 seconds due to the loading
      from .mainManager import MainManager
      from .utilizationListener import UtilizationListener
      from .optimizerListener import OptimizerListener
      from .datadogManager import DatadogManager

      ul = UtilizationListener()
      ol = OptimizerListener(n)
      ddg = DatadogManager()
      mm = MainManager(ddg, filter_tags)

      # utilization listeners
      if not optimize:
        mm.add_listener('ec2', ul.per_ec2)
        mm.add_listener('all', ul.after_all)
        mm.add_listener('all', ul.display_all)
      else:
        mm.add_listener('pre', ol.handle_pre)
        mm.add_listener('ec2', ol.per_ec2)
        mm.add_listener('all', ol.after_all)
        mm.add_listener('all', ol.storecsv_all)
        mm.add_listener('all', ol.display_all)


      # start download data and processing
      logger.info("Fetching history...")
      mm.get_ifi()

    except IsitfitError as e_info:
      logger.error("Error: %s"%str(e_info))
      import sys
      sys.exit(1)

    finally:
      display_footer()

#-----------------------

@cli.group(help="Explore EC2 tags", invoke_without_command=False)
def tags():
  pass



@tags.command(help="Generate new tags suggested by isitfit for each EC2 instance")
@click.option('--advanced', is_flag=True, help='Get advanced suggestions of tags. Requires login')
@click.pass_context
def suggest(ctx, advanced):
  # gather anonymous usage statistics
  from .utils import ping_matomo
  ping_matomo("/tags/suggest")

  from .utils import IsitfitError
  tl = None
  if not advanced:
    from .tagsSuggestBasic import TagsSuggestBasic
    tl = TagsSuggestBasic()
  else:
    from .tagsSuggestAdvanced import TagsSuggestAdvanced
    tl = TagsSuggestAdvanced()

  try:
    tl.prepare()
    tl.fetch()
    tl.suggest()
    tl.display()
  except IsitfitError as e:
    logger.error("Error: %s"%str(e))
    import sys
    sys.exit(1)

  display_footer()


@tags.command(help="Dump existing EC2 tags in tabular form into a csv file")
@click.pass_context
def dump(ctx):
  # gather anonymous usage statistics
  from .utils import ping_matomo
  ping_matomo("/tags/dump")

  from .tagsDump import TagsDump
  from .utils import IsitfitError
  tl = TagsDump()

  try:
    tl.fetch()
    tl.suggest() # not really suggesting. Just dumping to csv
    tl.display()
  except IsitfitError as e:
    logger.error("Error: %s"%str(e))
    import sys
    sys.exit(1)

  display_footer()



@tags.command(help="Push EC2 tags from csv file")
@click.argument('csv_filename') #, help='Path to CSV file holding tags to be pushed. Should match format from `isitfit tags dump`')
@click.option('--not-dry-run', is_flag=True, help='True for dry run (simulated push)')
def push(csv_filename, not_dry_run):
  # gather anonymous usage statistics
  from .utils import ping_matomo
  ping_matomo("/tags/push")

  from .tagsPush import TagsPush
  from .utils import IsitfitError

  tp = TagsPush(csv_filename)
  try:
    tp.read_csv()
    tp.validateTagsFile()
    tp.pullLatest()
    tp.diffLatest()
    tp.processPush(not not_dry_run)
  except IsitfitError as e:
    logger.error("Error: %s"%str(e))
    import sys
    sys.exit(1)

  display_footer()



#-----------------------

if __name__ == '__main__':
  cli()



from .utils import IsitfitError

import logging
logger = logging.getLogger('isitfit')

# https://pypi.org/project/termcolor/
from termcolor import colored

class TagsCsvDiff:

  def __init__(self, df_old, df_new):
    """
    df_old - pandas dataframe representing old tags
    df_new - new tags (pd dataframe)
    """
    self.df_old = df_old
    self.df_new = df_new
    self.old_minus_new = set()
    self.new_minus_old = set()
    self.migrations = []

  def noChanges(self):
    """
    Fast way to identify that no changes
    """
    to_json = lambda x: x.sort_values('instance_id', ascending=True)[sorted(list(x.columns))].to_json(orient='records')
    json_old = to_json(self.df_old)
    json_new = to_json(self.df_new)
    if json_old == json_new:
      raise IsitfitError("Aborting `tags push` due to no changes detected.")

  def noNewInstances(self):
    inst_old = set(self.df_old.instance_id)
    inst_new = set(self.df_new.instance_id)
    inst_created = inst_new - inst_old
    inst_created = sorted(list(inst_created))
    if len(inst_created)>0:
      msg_1 = "Found new instances IDs: %s%s"
      msg_2 = ", ".join(inst_created[:5])
      msg_3 ="..." if len(inst_created)>5 else ""
      msg_4 = msg_1%(msg_2,msg_3)
      raise IsitfitError(msg_4)


  def getDiffCols(self):
    # diff columns
    old_cols = set(self.df_old.columns)
    new_cols = set(self.df_new.columns)
    self.old_minus_new = old_cols - new_cols
    self.new_minus_old = new_cols - old_cols

  def renamedTags(self):
    # calculate string distances of all differences
    # https://stackoverflow.com/a/31236578/4126114
    # https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.corr.html
    import difflib
    corr_method = lambda s1, s2: 1 if s1 in s2 else (difflib.SequenceMatcher(None, s1, s2).ratio())
    import pandas as pd
    import numpy as np
    col_dist = pd.DataFrame(
      np.zeros(shape=(
          len(self.old_minus_new),
          len(self.new_minus_old)
        )),
      index=self.old_minus_new,
      columns=self.new_minus_old
    )
    new_processed = set()
    old_processed = set()
    for c1 in self.old_minus_new:
      for c2 in self.new_minus_old:
        d12 = corr_method(c1, c2)
        col_dist.loc[c1, c2] = d12

      # find best match
      d1_maxD = col_dist.loc[c1].max()
      if d1_maxD > 0.7:
        d1_maxV = col_dist.loc[c1].idxmax()
        confirm_msg = colored('Did you rename the tag "%s" to "%s"? yes/[no] '%(c1, d1_maxV), 'cyan')
        confirm_mv = input(confirm_msg)
        if confirm_mv.lower() in ['y', 'yes']:
          mig_i = ('mv', c1, d1_maxV)
          self.migrations.append(mig_i)
          old_processed.add(c1)
          new_processed.add(d1_maxV)

    # remove processed entries
    self.new_minus_old -= new_processed
    self.old_minus_new -= old_processed


  def newTags(self):
    """
    Identify if some tags are completely new.
    """
    if len(self.new_minus_old)==0:
      return

    #msg = []
    #msg.append("The following tags are completely new:")
    #msg = msg + sorted(list(self.new_minus_old))
    #msg = "\n".join(msg)
    #logger.info(msg)
    #logger.info("")

    logger.info("Found %i new tag(s)"%len(self.new_minus_old))
    logger.info("")

    new_processed = set()
    for ni in self.new_minus_old:
      confirm_msg = colored('Did you add the tag "%s"? yes/[no] '%ni, 'cyan')
      confirm_new = input(confirm_msg)
      if confirm_new.lower() in ['y', 'yes']:
        mig_i = ('touch', None, ni)
        self.migrations.append(mig_i)
        new_processed.add(ni)

    # remove all processed
    self.new_minus_old -= new_processed

  def droppedTags(self):
    """
    Identify if some tags are completely dropped.
    Ask the user if indeed dropped, or accident.
    Follows the idea of django/db/migrations/questioner.py
    where django asks if fields are renamed or dropped
    https://github.com/django/django/blob/e90af8bad44341cf8ebd469dac57b61a95667c1d/django/db/migrations/questioner.py
    """
    if len(self.old_minus_new)==0:
      return

    #msg = []
    #msg.append("The following tags are no longer present:")
    #msg = msg + sorted(list(self.old_minus_new))
    #msg = "\n".join(msg)
    #logger.info(msg)
    #logger.info("")

    logger.info("There are %i deleted tag(s)"%len(self.old_minus_new))
    logger.info("")

    old_processed = set()
    for ni in self.old_minus_new:
      confirm_msg = colored('Did you completely delete the tag "%s"? yes/[no] '%ni, 'cyan')
      confirm_del = input(confirm_msg)
      if confirm_del.lower() in ['y', 'yes']:
        mig_i = ('rm', ni, None)
        self.migrations.append(mig_i)
        old_processed.add(ni)

    # remove all processed
    self.old_minus_new -= old_processed


  def anyRemaining(self):
    if len(self.new_minus_old)>0:
      msg_1 = "Aborting `tags push` due to new tags in pushed csv file, and user indicated they shouldnt be created"
      msg_2 = ", ".join(sorted(list(self.new_minus_old)))
      msg_3 = "%s: %s"%(msg_1,msg_2)
      raise IsitfitError(msg_3)

    if len(self.old_minus_new)>0:
      msg_1 = "Aborting `tags push` due to missing tags in pushed csv file, and user indicated they shouldnt be deleted"
      msg_2 = ", ".join(sorted(list(self.old_minus_new)))
      msg_3 = "%s: %s"%(msg_1,msg_2)
      raise IsitfitError(msg_3)
